package sos.net;


import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import sos.net.SOSFTP;
import sos.net.SOSMail;

import sos.settings.SOSProfileSettings;
import sos.settings.SOSSettings;

import sos.util.SOSArguments;
import sos.util.SOSFile;
import sos.util.SOSFileOperations;
import sos.util.SOSGZip;
import sos.util.SOSLogger;
import sos.util.SOSStandardLogger;
import sos.util.SOSString;

/**
 * FTP File Transfer
 *
 * @author Andreas Pschel 
 */

public class SOSFTPCommand {
    
    /** The FTP server will always reply the ftp error codes, 
     * see http://www.the-eggman.com/seminars/ftp_error_codes.html */
    public  static final int ERROR_CODE              = 300; 
    
    private SOSString        sosString               = new SOSString();
    
    private SOSLogger        logger                  = null;
    
    private String           settingsSection         = "";
    
    private String           settingsFile            = "";
    
    private SOSSettings      settings                = null;
    
    public boolean           state                   = false;
    
    public String            stateText               = "";
    
    public String            mailSMTP                = "localhost";
    
    public String            mailFrom                = "SOSFTPCommand";
    
    public String            mailQueueDir            = "";
    
    
    
    public SOSFTPCommand(String settingsFile) {
        
        this.setSettingsFile(settingsFile);
    }
    
    public SOSFTPCommand(String settingsFile, SOSLogger logger) {
        
        this.setSettingsFile(settingsFile);
        this.setLogger(logger);
    }
    
    
    public SOSFTPCommand(String settingsFile, String settingsSection) {
        
        this.setSettingsFile(settingsFile);
        this.setSettingsSection(settingsSection);
    }
    
    
    public SOSFTPCommand(String settingsFile, String settingsSection, SOSLogger logger) {
        
        this.setSettingsFile(settingsFile);
        this.setSettingsSection(settingsSection);
        this.setLogger(logger);
    }
    
    
    
    
    public boolean receive() throws Exception {
        
        boolean      rc                              = false;       
        SOSFTP       ftpClient                       = null;                
        
        
        String       host                            = "";
        int          port                            = 21;
        
        String       user                            = "anonymous";
        String       password                        = "";
        String       account                         = "";
        
        String       transferMode                    = "binary";
        boolean      passiveMode                     = false;
        
        String       remoteDir                       = "./";
        String       localDir                        = ".";
        String       fileSpec                        = ".*";
        String       atomicSuffix                    = "";
        boolean      checkSize                       = true;
        long         checkInterval                   = 60;
        long         checkRetry                      = 0;
        boolean      overwriteFiles                  = true;
        boolean      appendFiles                     = false;
        boolean      removeFiles                     = false;
        boolean      forceFiles                      = true;
        
        boolean      zeroByteFiles                   = true;
        boolean      zeroByteFilesStrict             = false; 
        boolean      zeroByteFilesRelaxed            = false;
        
        String       fileNotificationTo              = "";
        String       fileNotificationCC              = "";
        String       fileNotificationBCC             = "";
        String       fileNotificationSubject         = "";
        String       fileNotificationBody            = "";
        
        String       fileZeroByteNotificationTo      = "";
        String       fileZeroByteNotificationCC      = "";
        String       fileZeroByteNotificationBCC     = "";
        String       fileZeroByteNotificationSubject = "";
        String       fileZeroByteNotificationBody    = "";
        
        boolean      isLoggedIn                      = false;
        Vector       transferFileList                = new Vector();
        HashMap      checkFileList                   = new HashMap();
        
        String       replacing                       = null;
        String       replacement                     = null;    
        
        boolean      recursive                       = false;
        
        int          pollTimeout                     = 0;
        int          pollIntervall                   = 60;
        int          pollMinFiles                    = 1;
        
        String       filePath                        = "";
        boolean      isFilePath                      = false;
        
//      alternative Parameter;
        String       alternativeHost                 = "";
        int          alternativePort                 = 0;
        String       alternativeUser                 = ""; 
        String       alternativePassword             = "";
        String       alternativeAccount              = "";
        String       alternativeRemoteDir            = "";
        boolean      alternativePassiveMode          = false;
        String       alternativeTransferMode         = "binary";
        
        try {   
            if (this.getLogger() == null) this.setLogger(new SOSStandardLogger(0));
            
            try {
                this.setSettings(new SOSProfileSettings( this.getSettingsFile(), this.getSettingsSection(), this.getLogger()));
            } catch (Exception e) {
                throw new Exception("error occurred retrieving settings: " + e.getMessage());
            }
            
            try { // to get the parameters
                if (settings.getSectionEntry("ftp_host") != null && settings.getSectionEntry("ftp_host").length() > 0)
                    host = settings.getSectionEntry("ftp_host");
                
                if (settings.getSectionEntry("ftp_port") != null && settings.getSectionEntry("ftp_port").length() > 0) {
                    try {
                        port = Integer.parseInt(settings.getSectionEntry("ftp_port"));
                    } catch (Exception e) {
                        throw new Exception("illegal value for parameter [ftp_port]: " + settings.getSectionEntry("ftp_port"));
                    }
                }
                
                if (settings.getSectionEntry("ftp_user") != null && settings.getSectionEntry("ftp_user").length() > 0)
                    user = settings.getSectionEntry("ftp_user");
                if (settings.getSectionEntry("ftp_password") != null && settings.getSectionEntry("ftp_password").length() > 0)
                    password = settings.getSectionEntry("ftp_password");
                if (settings.getSectionEntry("ftp_account") != null && settings.getSectionEntry("ftp_account").length() > 0)
                    account = settings.getSectionEntry("ftp_account");
                
                if (settings.getSectionEntry("ftp_transfer_mode") != null && settings.getSectionEntry("ftp_transfer_mode").length() > 0)
                    transferMode = settings.getSectionEntry("ftp_transfer_mode");
                
                if (settings.getSectionEntry("ftp_passive_mode") != null && settings.getSectionEntry("ftp_passive_mode").length() > 0)
                    passiveMode = (settings.getSectionEntry("ftp_passive_mode").equals("1") ||
                            settings.getSectionEntry("ftp_passive_mode").equalsIgnoreCase("true") ||
                            settings.getSectionEntry("ftp_passive_mode").equalsIgnoreCase("yes")? true : false);
                
                if (settings.getSectionEntry("ftp_remote_dir") != null && settings.getSectionEntry("ftp_remote_dir").length() > 0)
                    remoteDir = settings.getSectionEntry("ftp_remote_dir");
                
                if (settings.getSectionEntry("ftp_local_dir") != null && settings.getSectionEntry("ftp_local_dir").length() > 0)
                    localDir = settings.getSectionEntry("ftp_local_dir");
                
                if (settings.getSectionEntry("ftp_file_spec") != null && settings.getSectionEntry("ftp_file_spec").length() > 0)
                    fileSpec = settings.getSectionEntry("ftp_file_spec");
                
                if (settings.getSectionEntry("ftp_atomic_suffix") != null && settings.getSectionEntry("ftp_atomic_suffix").length() > 0)
                    atomicSuffix = settings.getSectionEntry("ftp_atomic_suffix");
                
                
                if (settings.getSectionEntry("ftp_check_size") != null && settings.getSectionEntry("ftp_check_size").length() > 0)
                    checkSize = (!settings.getSectionEntry("ftp_check_size").equals("1") 
                            && !settings.getSectionEntry("ftp_check_size").equalsIgnoreCase("true") 
                            && !settings.getSectionEntry("ftp_check_size").equalsIgnoreCase("yes") ? false : true);
                
                if (settings.getSectionEntry("ftp_check_interval") != null && settings.getSectionEntry("ftp_check_interval").length() > 0) {
                    try {
                        checkInterval = Long.parseLong(settings.getSectionEntry("ftp_check_interval"));
                    } catch (Exception e) {
                        throw new Exception("illegal value for parameter [ftp_check_interval]: " + settings.getSectionEntry("ftp_check_interval"));
                    }
                }
                
                if (settings.getSectionEntry("ftp_check_retry") != null && settings.getSectionEntry("ftp_check_retry").length() > 0) {
                    try {
                        checkRetry = Long.parseLong(settings.getSectionEntry("ftp_check_retry"));
                    } catch (Exception e) {
                        throw new Exception("illegal value for parameter [ftp_check_retry]: " + settings.getSectionEntry("ftp_check_retry"));
                    }
                }
                
                if (settings.getSectionEntry("ftp_overwrite_files") != null && settings.getSectionEntry("ftp_overwrite_files").length() > 0)
                    overwriteFiles = (!settings.getSectionEntry("ftp_overwrite_files").equals("1") 
                            && !settings.getSectionEntry("ftp_overwrite_files").equalsIgnoreCase("true") 
                            && !settings.getSectionEntry("ftp_overwrite_files").equalsIgnoreCase("yes") ? false : true);
                
                if (settings.getSectionEntry("ftp_append_files") != null && settings.getSectionEntry("ftp_append_files").length() > 0)
                    appendFiles = (settings.getSectionEntry("ftp_append_files").equals("1") 
                            || settings.getSectionEntry("ftp_append_files").equalsIgnoreCase("true") 
                            || settings.getSectionEntry("ftp_append_files").equalsIgnoreCase("yes") ? true : false);
                
                if (settings.getSectionEntry("ftp_remove_files") != null && settings.getSectionEntry("ftp_remove_files").length() > 0)
                    removeFiles = (settings.getSectionEntry("ftp_remove_files").equals("1") 
                            || settings.getSectionEntry("ftp_remove_files").equalsIgnoreCase("true") 
                            || settings.getSectionEntry("ftp_remove_files").equalsIgnoreCase("yes") ? true : false);
                
                if (settings.getSectionEntry("ftp_force_files") != null && settings.getSectionEntry("ftp_force_files").length() > 0)
                    forceFiles = (!settings.getSectionEntry("ftp_force_files").equals("1") 
                            && !settings.getSectionEntry("ftp_force_files").equalsIgnoreCase("true") 
                            && !settings.getSectionEntry("ftp_force_files").equalsIgnoreCase("yes") ? false : true);
                
                
                if (settings.getSectionEntry("ftp_file_zero_byte_transfer") != null && settings.getSectionEntry("ftp_file_zero_byte_transfer").length() > 0) {
                    if (settings.getSectionEntry("ftp_file_zero_byte_transfer").equals("1") 
                            || settings.getSectionEntry("ftp_file_zero_byte_transfer").equalsIgnoreCase("true") 
                            || settings.getSectionEntry("ftp_file_zero_byte_transfer").equalsIgnoreCase("yes")) {
                        zeroByteFiles = true;
                        zeroByteFilesStrict = false;
                    } else if (settings.getSectionEntry("ftp_file_zero_byte_transfer").equalsIgnoreCase("strict")) {
                        zeroByteFiles = false;
                        zeroByteFilesStrict = true;
                    } else if (settings.getSectionEntry("ftp_file_zero_byte_transfer").equalsIgnoreCase("relaxed")) {
                        zeroByteFiles = false;
                        zeroByteFilesStrict = false;
                        zeroByteFilesRelaxed = true;
                    } else {
                        zeroByteFiles = false;
                        zeroByteFilesStrict = false;
                    }
                }
                
                if (settings.getSectionEntry("ftp_file_notification_to") != null && settings.getSectionEntry("ftp_file_notification_to").length() > 0) {
                    fileNotificationTo = settings.getSectionEntry("ftp_file_notification_to");
                }
                
                if (settings.getSectionEntry("ftp_file_notification_cc") != null && settings.getSectionEntry("ftp_file_notification_cc").length() > 0) {
                    fileNotificationCC = settings.getSectionEntry("ftp_file_notification_cc");
                }
                
                if (settings.getSectionEntry("ftp_file_notification_bcc") != null && settings.getSectionEntry("ftp_file_notification_bcc").length() > 0) {
                    fileNotificationBCC = settings.getSectionEntry("ftp_file_notification_bcc");
                }
                
                if (settings.getSectionEntry("ftp_file_notification_subject") != null && settings.getSectionEntry("ftp_file_notification_subject").length() > 0) {
                    fileNotificationSubject = settings.getSectionEntry("ftp_file_notification_subject");
                }
                
                if (settings.getSectionEntry("ftp_file_notification_body") != null && settings.getSectionEntry("ftp_file_notification_body").length() > 0) {
                    fileNotificationBody = settings.getSectionEntry("ftp_file_notification_body");
                }
                
                
                if (settings.getSectionEntry("ftp_file_zero_byte_notification_to") != null && settings.getSectionEntry("ftp_file_zero_byte_notification_to").length() > 0) {
                    fileZeroByteNotificationTo = settings.getSectionEntry("ftp_file_zero_byte_notification_to");
                }
                
                if (settings.getSectionEntry("ftp_file_zero_byte_notification_cc") != null && settings.getSectionEntry("ftp_file_zero_byte_notification_cc").length() > 0) {
                    fileZeroByteNotificationCC = settings.getSectionEntry("ftp_file_zero_byte_notification_cc");
                }
                
                if (settings.getSectionEntry("ftp_file_zero_byte_notification_bcc") != null && settings.getSectionEntry("ftp_file_zero_byte_notification_bcc").length() > 0) {
                    fileZeroByteNotificationBCC = settings.getSectionEntry("ftp_file_zero_byte_notification_bcc");
                }
                
                if (settings.getSectionEntry("ftp_file_zero_byte_notification_subject") != null && settings.getSectionEntry("ftp_file_zero_byte_notification_subject").length() > 0) {
                    fileZeroByteNotificationSubject = settings.getSectionEntry("ftp_file_zero_byte_notification_subject");
                }
                
                if (settings.getSectionEntry("ftp_file_zero_byte_notification_body") != null && settings.getSectionEntry("ftp_file_zero_byte_notification_body").length() > 0) {
                    fileZeroByteNotificationBody = settings.getSectionEntry("ftp_file_zero_byte_notification_body");
                }
                
                if ( settings.getSectionEntry("replacing")!=null && settings.getSectionEntry("replacing").length()>0 ) {                
                    replacing = settings.getSectionEntry("replacing");                  
                }
                
                if ( settings.getSectionEntry("replacement")!=null && settings.getSectionEntry("replacement").length()>0 ) {                
                    replacement = settings.getSectionEntry("replacement");                  
                }    
                
                if ( settings.getSectionEntry("ftp_recursive")!=null && settings.getSectionEntry("ftp_recursive").length()>0 ) {    
                    String sRecursive = "";
                    sRecursive = settings.getSectionEntry("ftp_recursive");
                    recursive = sosString.parseToBoolean(sRecursive);
                }  
                
                if ( replacing != null && replacement == null ) {
                    throw new Exception("job parameter is missing for specified parameter [replacing]: [replacement]");
                }
                
                if ( replacing == null && replacement != null ) {
                    throw new Exception("job parameter is missing for specified parameter [replacement]: [replacing]");
                }
                
                if ( settings.getSectionEntry("ftp_poll_timeout")!=null && settings.getSectionEntry("ftp_poll_timeout").length()>0 ) {              
                    pollTimeout = Integer.parseInt(settings.getSectionEntry("ftp_poll_timeout"));                   
                }
                
                if ( settings.getSectionEntry("ftp_poll_interval")!=null && settings.getSectionEntry("ftp_poll_interval").length()>0 ) {                
                    pollIntervall = Integer.parseInt(settings.getSectionEntry("ftp_poll_interval"));                    
                }
                
                if ( settings.getSectionEntry("ftp_poll_minfiles")!=null && settings.getSectionEntry("ftp_poll_minfiles").length()>0 ) {                
                    pollMinFiles = Integer.parseInt(settings.getSectionEntry("ftp_poll_minfiles"));                 
                }
                
//              alternative Parameter
                if (settings.getSectionEntry("ftp_alternative_host") != null && settings.getSectionEntry("ftp_alternative_host").length() > 0) 
                    alternativeHost = settings.getSectionEntry("ftp_alternative_host");                   
                
                if (settings.getSectionEntry("ftp_alternative_port") != null && settings.getSectionEntry("ftp_alternative_port").length() > 0)
                    alternativePort = Integer.parseInt(settings.getSectionEntry("ftp_alternative_port"));
                
                if (settings.getSectionEntry("ftp_alternative_password") != null && settings.getSectionEntry("ftp_alternative_password").length() > 0)
                    alternativePassword = settings.getSectionEntry("ftp_alternative_password");
                
                if (settings.getSectionEntry("ftp_alternative_user") != null && settings.getSectionEntry("ftp_alternative_user").length() > 0)
                    alternativeUser = settings.getSectionEntry("ftp_alternative_user");
                
                if (settings.getSectionEntry("ftp_alternative_account") != null && settings.getSectionEntry("ftp_alternative_account").length() > 0)
                    alternativeAccount = settings.getSectionEntry("ftp_alternative_account");
                
                if (settings.getSectionEntry("ftp_alternative_remote_dir") != null && settings.getSectionEntry("ftp_alternative_remote_dir").length() > 0)
                    alternativeRemoteDir = settings.getSectionEntry("ftp_alternative_remote_dir");
                
                if (settings.getSectionEntry("ftp_alternative_passive_mode") != null && settings.getSectionEntry("ftp_alternative_passive_mode").length() > 0)
                    alternativePassiveMode = (settings.getSectionEntry("ftp_alternative_passive_mode").equals("1") ? true : false);
                
                if (settings.getSectionEntry("ftp_alternative_transfer_mode") != null && settings.getSectionEntry("ftp_alternative_transfer_mode").length() > 0)
                    alternativeTransferMode = settings.getSectionEntry("ftp_alternative_transfer_mode");
                
            } catch (Exception e) {
                rc = false;
                throw (new Exception("could not process job parameters: " + e.getMessage()));
            }
            
            
            if (fileNotificationTo != null && fileNotificationTo.length() > 0) {
                if (fileNotificationSubject == null || fileNotificationSubject.length() == 0) {
                     fileNotificationSubject = "SOSFTPCommand";
                }
                
                if (fileNotificationBody == null || fileNotificationBody.length() == 0) {
                    fileNotificationBody = "The following files have been received:\n\n";
                }
            }
            
            if (fileZeroByteNotificationTo != null && fileZeroByteNotificationTo.length() > 0) {
                if (fileZeroByteNotificationSubject == null || fileZeroByteNotificationSubject.length() == 0) {
                    fileZeroByteNotificationSubject = "[warning] SOSFTPCommand";
                }
                
                if (fileZeroByteNotificationBody == null || fileZeroByteNotificationBody.length() == 0) {
                    fileZeroByteNotificationBody = "The following files have been received and were removed due to zero byte constraints:\n\n";
                }
            }           
            
            if (settings.getSectionEntry("ftp_file_path") != null && settings.getSectionEntry("ftp_file_path").length() > 0) {
                filePath = settings.getSectionEntry("ftp_file_path");
                isFilePath = true;
            } else {
                isFilePath = false;
            }
            
            this.getLogger().debug(".. job parameter [ftp_host]                               : " + host);
            this.getLogger().debug(".. job parameter [ftp_port]                               : " + port);
            this.getLogger().debug(".. job parameter [ftp_user]                               : " + user);          
            this.getLogger().debug(".. job parameter [ftp_account]                            : " + account);
            this.getLogger().debug(".. job parameter [ftp_transfer_mode]                      : " + transferMode);
            this.getLogger().debug(".. job parameter [ftp_passive_mode]                       : " + passiveMode);
            this.getLogger().debug(".. job parameter [ftp_remote_dir]                         : " + remoteDir);
            this.getLogger().debug(".. job parameter [ftp_local_dir]                          : " + localDir);
            this.getLogger().debug(".. job parameter [ftp_file_spec]                          : " + fileSpec);
            this.getLogger().debug(".. job parameter [ftp_atomic_suffix]                      : " + atomicSuffix);
            this.getLogger().debug(".. job parameter [ftp_check_size]                         : " + checkSize);
            this.getLogger().debug(".. job parameter [ftp_check_interval]                     : " + checkInterval);
            this.getLogger().debug(".. job parameter [ftp_check_retry]                        : " + checkRetry);
            this.getLogger().debug(".. job parameter [ftp_overwrite_files]                    : " + overwriteFiles);
            this.getLogger().debug(".. job parameter [ftp_append_files]                       : " + appendFiles);
            this.getLogger().debug(".. job parameter [ftp_remove_files]                       : " + removeFiles);
            this.getLogger().debug(".. job parameter [ftp_force_files]                        : " + forceFiles);
            
            this.getLogger().debug(".. job parameter [ftp_zero_byte_transfer] zeroByte        : " + zeroByteFiles);
            this.getLogger().debug(".. job parameter [ftp_zero_byte_transfer] strict          : " + zeroByteFilesStrict);
            this.getLogger().debug(".. job parameter [ftp_zero_byte_transfer] relaxed         : " + zeroByteFilesRelaxed);
            
            this.getLogger().debug(".. job parameter [ftp_file_notification_to]               : " + fileNotificationTo);
            this.getLogger().debug(".. job parameter [ftp_file_notification_cc]               : " + fileNotificationCC);
            this.getLogger().debug(".. job parameter [ftp_file_notification_bcc]              : " + fileNotificationBCC);
            this.getLogger().debug(".. job parameter [ftp_file_notification_subject]          : " + fileNotificationSubject);
            this.getLogger().debug(".. job parameter [ftp_file_notification_body]             : " + fileNotificationBody);
            
            this.getLogger().debug(".. job parameter [ftp_file_zero_byte_notification_to]     : " + fileZeroByteNotificationTo);
            this.getLogger().debug(".. job parameter [ftp_file_zero_byte_notification_cc]     : " + fileZeroByteNotificationCC);
            this.getLogger().debug(".. job parameter [ftp_file_zero_byte_notification_bcc]    : " + fileZeroByteNotificationBCC);
            this.getLogger().debug(".. job parameter [ftp_file_zero_byte_notification_subject]: " + fileZeroByteNotificationSubject);
            this.getLogger().debug(".. job parameter [ftp_file_zero_byte_notification_body]   : " + fileZeroByteNotificationBody);
            
            this.getLogger().debug(".. job parameter [replacing]                              : " + replacing);
            this.getLogger().debug(".. job parameter [replacement]                            : " + replacement);
            
            this.getLogger().debug(".. job parameter [ftp_recursive]                          : " + recursive);
            
            this.getLogger().debug(".. job parameter [ftp_poll_timeout]                       : " + pollTimeout);
            this.getLogger().debug(".. job parameter [ftp_poll_interval]                      : " + pollIntervall);
            this.getLogger().debug(".. job parameter [ftp_poll_minfiles]                      : " + pollMinFiles);          
            
            this.getLogger().debug(".. job parameter [ftp_file_path]                          : " + filePath);
            
            //alternative Parameter
            this.getLogger().debug(".. job parameter [ftp_alternative_host]                   : " + alternativeHost);
            this.getLogger().debug(".. job parameter [ftp_alternative_port]                   : " + alternativePort);
            this.getLogger().debug(".. job parameter [ftp_alternative_user]                   : " + alternativeUser);
            this.getLogger().debug(".. job parameter [ftp_alternative_password]               : " + alternativePassword);
            this.getLogger().debug(".. job parameter [ftp_alternative_account]                : " + alternativeAccount);
            this.getLogger().debug(".. job parameter [ftp_alternative_remote_dir]             : " + alternativeRemoteDir);
            this.getLogger().debug(".. job parameter [ftp_alternative_passive_mode]           : " + alternativePassiveMode);
            this.getLogger().debug(".. job parameter [ftp_alternative_transfer_mode]          : " + alternativeTransferMode);
            
            
            try { // to check parameters
                if (host == null || host.length() == 0) throw new Exception("no host was specified");
                if (user == null || user.length() == 0) throw new Exception("no user was specified");                   
                if(sosString.parseToString(filePath).length() > 0) {
                    if(pollMinFiles > 1) {
                        getLogger().warn("parameter [ftp_poll_minfiles] should not be greater than 1 because Parameter ftp_file_path has been specified. ");                      
                    }                
                }
            } catch (Exception e) {
                rc = false;
                throw (new Exception("invalid or insufficient parameters: " + e.getMessage()));
            }
            
            
            try { // to process ftp
                this.getLogger().info("connecting to host " + host + ", port " + port + ", local directory " + localDir + ", remote directory " + remoteDir + 
                        (isFilePath ? ", file " + filePath : ", file specification " + fileSpec) );
                
                boolean alternativeUse = true;//Wiederholungskriterien
                int isAlternativeParameterUse = 0;//Abbruchbedingug, versuch mit alternative Parameter nur einmal
                while (alternativeUse && isAlternativeParameterUse <= 1) {
                    try {
                        ftpClient = new SOSFTP(host, port);
                        this.getLogger().debug("..ftp server reply [init] [host=" + host + "], [port="+ port + "]: " + ftpClient.getReplyString() );
                        
                        if (account != null && account.length() > 0) {
                            isLoggedIn = ftpClient.login(user, password, account);
                            this.getLogger().debug("..ftp server reply [login] [user=" + user + "], [account=" + account + "]: " + ftpClient.getReplyString() );
                        } else {
                            isLoggedIn = ftpClient.login(user, password);
                            this.getLogger().debug("..ftp server reply [login] [user=" + user + "]: " + ftpClient.getReplyString() );
                        }
                        
                        if (!isLoggedIn || ftpClient.getReplyCode() > ERROR_CODE) {
                            throw new Exception("..ftp server reply [login failed] [user=" + user + "], [account=" + account + "]: " + ftpClient.getReplyString() );
                        }
                        alternativeUse = false;
                    } catch (Exception ex) {
                        
                        this.getLogger().info("..error in ftp server init with [host=" + host + "], [port="+ port + "] " + ex.getMessage());
                        
                        alternativeUse = (alternativeHost.concat(alternativeUser).concat(alternativePassword).
                                concat(alternativeAccount).length() > 0 || alternativePort != 0);
                        
                        if(alternativeUse && isAlternativeParameterUse == 0) {
                            
                            if (ftpClient != null) {
                                if (isLoggedIn) try { ftpClient.logout(); } catch (Exception e) {} // no error handling
                                if (ftpClient.isConnected()) try { ftpClient.disconnect(); } catch(Exception e) {} // no error handling
                            }
                            
                            isAlternativeParameterUse++;
                            host = getAlternative(host, alternativeHost);
                            port = getAlternative(port, alternativePort);
                            user = getAlternative(user, alternativeUser); 
                            password = getAlternative(password, alternativePassword);                       
                            account = getAlternative(account, alternativeAccount);
                            remoteDir = getAlternative(remoteDir, alternativeRemoteDir);
                            passiveMode = alternativePassiveMode;
                            transferMode = getAlternative(transferMode, alternativeTransferMode);
                            
                            this.getLogger().info("..try login with alternative parameter [host=" + host + "], [port="+ port + "] " +
                                    "[user=" + user + "], [account=" + account + "], [remoteDir=" + remoteDir +"], [passiveMode=" + passiveMode +"], " +
                                    "[transferMode=" + transferMode + "]");
                            
                        } else {
                            //wie gehabt Fehler
                            throw new Exception ("..error in ftp server init with [host=" + host + "], [port="+ port + "] " + ex.getMessage());
                        }                           
                        
                    }
                }
                
                
                if (passiveMode) {
                    ftpClient.passive();
                    if (ftpClient.getReplyCode() > ERROR_CODE) {
                        throw new Exception("..ftp server reply [passive]: "  + ftpClient.getReplyString());
                    } else {
                        this.getLogger().debug("..ftp server reply [passive]: "  + ftpClient.getReplyString());
                    }                   
                }                           
                
                
                if (transferMode.equalsIgnoreCase("ascii")) {
                    if (ftpClient.ascii()) {
                        this.getLogger().debug("..using ASCII mode for file transfer");
                        this.getLogger().debug("..ftp server reply [ascii]: "  + ftpClient.getReplyString());
                    } else {
                        throw new Exception(".. could not switch to ASCII mode for file transfer ..ftp server reply [ascii]: "  + ftpClient.getReplyString());
                    }
                } else {
                    if (ftpClient.binary()) {                       
                        this.getLogger().debug("using binary mode for file transfers.");
                        this.getLogger().debug("..ftp server reply [binary]: "  + ftpClient.getReplyString());
                    } else {
                        throw new Exception(".. could not switch to binary mode for file transfer ..ftp server reply [ascii]: "  + ftpClient.getReplyString());
                    }
                }
                
                
                if(isFilePath) {
                    String currRemoteDir = new File(filePath).getParent().replaceAll("\\\\", "/");
                    if(sosString.parseToString(currRemoteDir).length() > 0) {
                        ftpClient.cd(currRemoteDir);
                        if (ftpClient.getReplyCode() > ERROR_CODE) {
                            throw new Exception("..ftp server reply [cd] [directory ftp_file_path=" + currRemoteDir + "]: "  + ftpClient.getReplyString());
                        } else {
                            getLogger().debug("..ftp server reply [cd] [directory ftp_file_path=" + currRemoteDir +"]: "  + ftpClient.getReplyString() );
                        }
                    }
                } else if (sosString.parseToString(remoteDir).length() > 0) {
                    ftpClient.cd(remoteDir);
                    if (ftpClient.getReplyCode() > ERROR_CODE) {
                        throw new Exception("..ftp server reply [cd] [remoteDir=" + remoteDir + "]: "  + ftpClient.getReplyString());
                    } else {
                        getLogger().debug("..ftp server reply [cd] [remoteDir=" + remoteDir + "]: "  + ftpClient.getReplyString() );
                    }                   
                }
                
                
                Vector filelist = null;
                
                if (isFilePath) {
                    this.getLogger().info("local directory: " + localDir + ", remote file: " + filePath);
                    filelist = new Vector();                    
                    filelist.add(filePath);     
                    fileSpec = ".*";
                } else {
                    this.getLogger().info("local directory: " + localDir + ", remote directory: " + remoteDir + ", file specification: " + fileSpec);
                    filelist = ftpClient.nList(recursive);
                }
                int count = 0;
                int zeroByteCount = 0;
                
                if (ftpClient.getReplyCode() > ERROR_CODE) {
                    // no error checking for empty directories
                    // throw new Exception("..error occurred looking up remote files [" + remoteDir + "]: "  + ftpClient.getReplyString());
                } else {
                    getLogger().debug("..ftp server reply [nlist]: " + ftpClient.getReplyString() );
                }
                
                Pattern pattern = Pattern.compile(fileSpec, 0);
                Iterator iterator = filelist.iterator();
                
                if (pollTimeout>0){
                    // before any processing, check if files are there
                    boolean done= false;
                    boolean giveUpPoll = false;
                    double delay = pollIntervall;
                    double nrOfTries = (pollTimeout*60)/delay;
                    int tries=0;
                    while(!done && !giveUpPoll){                        
                        tries++;
                        int matchedFiles=0;                     
                        while(iterator.hasNext()) {
                            String fileName = (String)iterator.next();
                            if (isFilePath) {
                                matchedFiles++;                             
                            } else {
                                Matcher matcher = pattern.matcher(fileName);
                                if (matcher.find()) matchedFiles++;
                            }
                        }
                        getLogger().debug3(matchedFiles +" matching files found");
                        if (matchedFiles < pollMinFiles){                           
                                if (tries<nrOfTries){
                                    Thread.sleep((long)delay * 1000);
                                    this.setStateText("Polling for files... ");
                                    if (isFilePath) {                   
                                        filelist = new Vector();
                                        filelist.add(filePath);                             
                                    } else {
                                        filelist = ftpClient.nList(recursive);
                                    }
                                    iterator = filelist.iterator();
                                } else {
                                    giveUpPoll = true;
                                }
                        } else{
                            done=true;
                        }
                        if (giveUpPoll){
                            String message = "Failed to find at least "+pollMinFiles+" files matching \""+ 
                            fileSpec +"\" after triggering for "+pollTimeout+" minutes.";
                            if (matchedFiles>0) message+=" (Only "+matchedFiles+" files were found)";
                            if (forceFiles) {
                                this.getLogger().warn(message);
                                String body = message +"\n";
                                body += "See attached logfile for details.";
                                // spooler_log.mail().set_body(body);
                            }else {
                                this.getLogger().info(message);
                                return true;
                            }
                            return false;
                        }
                    }                   
                }
                
                iterator = filelist.iterator();
                while(iterator.hasNext()) {
                    String fileName = (String)iterator.next();
                    Matcher matcher = pattern.matcher(fileName);
                    File transferFile = null; 
                    
                    if(recursive && !isFilePath) {
                        if (sosString.parseToString(remoteDir).length() > 0) {
                            ftpClient.cd(remoteDir);
                            if (ftpClient.getReplyCode() > ERROR_CODE) {
                                throw new Exception("..ftp server reply [cd] [remoteDir=" + remoteDir + "]: "  + ftpClient.getReplyString());
                            } else {
                                getLogger().debug("..ftp server reply [cd] [remoteDir=" + remoteDir + "]: "  + ftpClient.getReplyString() );
                            }
                        }
                    }
                    
                    if (matcher.find() || isFilePath) {
                        
                        if (replacement != null && replacement.length() > 0 && replacing != null && replacing.length() > 0) {
                            String currTransferFilename = SOSFileOperations.getReplacementFilename(fileName, replacing, replacement);
                            this.getLogger().info(" transfer file [" + fileName + "] is renamed to " + currTransferFilename);
                            if(isFilePath) {
                                transferFile = new File(localDir 
                                        + (localDir.endsWith("/") || localDir.endsWith("\\") ? "" : "/") 
                                        + new File(currTransferFilename).getName());
                            } else {
                                transferFile = new File(localDir 
                                        + (localDir.endsWith("/") || localDir.endsWith("\\") ? "" : "/") 
                                        + currTransferFilename);
                            }
                        } else {
                            if(isFilePath) {
                                transferFile = new File(localDir 
                                        + (localDir.endsWith("/") || localDir.endsWith("\\") ? "" : "/") 
                                        + new File(fileName).getName());
                            } else {
                                transferFile = new File(localDir 
                                        + (localDir.endsWith("/") || localDir.endsWith("\\") ? "" : "/") 
                                        + fileName);
                            }
                        }
                        
                        File transFile = transferFile;
                        if(transFile.getParent() != null && !transFile.getParentFile().exists()){                                                       
                            //Unterverzeichnisse mit bercksichtigen
                            if(transFile.getParentFile().mkdirs()) {                    
                                this.getLogger().info("..create parent directory [" + transFile.getParentFile() + "]");
                            } else {
                                throw new Exception ("..error occurred creating directory [" + transFile.getParentFile() + "]");
                            }
                        }
                        
                        if (!appendFiles && !overwriteFiles && transferFile.exists()) {
                            this.getLogger().info("..ftp transfer skipped for file [no overwrite]: " + transferFile.getName());
                            continue;
                        }
                        
                        long bytesSent = 0;
                        
                        if (!appendFiles && atomicSuffix != null && atomicSuffix.length() > 0) {
                            File atomicFile = new File(transferFile.getAbsolutePath() + atomicSuffix);
                            
                            File file = new File(fileName); 
                            if(recursive) {
                                if(file.getParent() != null && !isFilePath) {
                                    String[] splitParent = file.getParent().split("\\\\");
                                    for (int i=0; i<splitParent.length; i++) {
                                        if(sosString.parseToString(splitParent[i]).length() > 0) {
                                            ftpClient.cd(splitParent[i]);
                                            if (ftpClient.getReplyCode() > ERROR_CODE) {
                                                throw new Exception("..ftp server reply [cd] [remoteDir=" + remoteDir + "]: "  + ftpClient.getReplyString());
                                            } else {
                                                getLogger().debug("..ftp server reply [cd] [remoteDir=" + remoteDir + "]: "  + ftpClient.getReplyString() );
                                            }
                                        }
                                    }
                                }
                            }
                            
                            
                            bytesSent = this.transferFile(ftpClient, new File(file.getName()), atomicFile, checkRetry, checkInterval, appendFiles);
                            
                            if (transferFile.exists()) {
                                if (overwriteFiles) {
                                    if (!transferFile.delete()) {
                                        throw new Exception("could not overwrite local file: " + transferFile.getAbsolutePath());
                                    }
                                }
                            }
                            
                            if (bytesSent <= 0 && !zeroByteFiles && zeroByteFilesStrict) {
                                this.getLogger().info("removing local file : " + transferFile.getAbsolutePath() + " due to zero byte strict constraint");
                                if (!atomicFile.delete()) {
                                    throw new Exception("..error occurred, could not remove temporary file: " + atomicFile.getAbsolutePath());
                                }
                                fileZeroByteNotificationBody += transferFile.getName() + "\n";
                                zeroByteCount++;
                            } else if (bytesSent <= 0 && !zeroByteFiles) {
                                checkFileList.put(atomicFile, transferFile);
                            } else {
                                this.getLogger().info("renaming local temporary file : " + transferFile.getAbsolutePath());
                                if (!atomicFile.renameTo(transferFile)) {
                                    throw new Exception("could not rename temporary file [" + atomicFile.getCanonicalPath() + "] to: " + transferFile.getAbsolutePath());
                                }
                                fileNotificationBody += transferFile.getName() + "\n";
                                transferFileList.add(transferFile);
                                count++;
                            }
                        } else {
                            
                            File file = new File(fileName); 
                            if(recursive) {
                                if(file.getParent() != null) {
                                    String[] splitParent = file.getParent().split("\\\\");
                                    for (int i=0; i<splitParent.length; i++) {
                                        ftpClient.cd(splitParent[i]);
                                        if (ftpClient.getReplyCode() > ERROR_CODE) {
                                            throw new Exception("..ftp server reply [cd] [remoteDir=" + remoteDir + "]: "  + ftpClient.getReplyString());
                                        } else {
                                            getLogger().debug("..ftp server reply [cd] [remoteDir=" + remoteDir + "]: "  + ftpClient.getReplyString() );
                                        }
                                    }
                                }
                            }
                            
                            bytesSent = this.transferFile(ftpClient, new File(file.getName()), transferFile, checkRetry, checkInterval, appendFiles);
                            
                            if (bytesSent <= 0 && !zeroByteFiles && zeroByteFilesStrict) {
                                this.getLogger().info("removing local file : " + transferFile.getAbsolutePath() + " due to zero byte strict constraint");
                                if (!transferFile.delete()) {
                                    throw new Exception("..error occurred, could not remove temporary file: " + transferFile.getAbsolutePath());
                                }
                                fileZeroByteNotificationBody += transferFile.getName() + "\n";
                                zeroByteCount++;
                            } else if (bytesSent <= 0 && !zeroByteFiles) {
                                checkFileList.put(transferFile, transferFile);
                            } else {
                                fileNotificationBody += transferFile.getName() + "\n";
                                transferFileList.add(transferFile);
                                count++;
                            }
                        }
                        
                        
                        if (removeFiles) {
                            boolean ok = ftpClient.delete(new File(fileName).getName());                            
                            if (ftpClient.getReplyCode() > ERROR_CODE) {
                                throw new Exception("..error occurred removing remote file [" + transferFile.getName() + "]: "  + ftpClient.getReplyString());
                            } else if (!ok) {
                                throw new Exception("..error occurred, could not remove remote file [" + transferFile.getName() + "]: " + ftpClient.getReplyString());
                            } else {
                                this.getLogger().info("removing remote file: " + transferFile.getName());
                            }
                        }
                    }
                }
                
                
                if (!zeroByteFiles) {
                    Iterator checkFileListIterator = checkFileList.keySet().iterator();
                    while(checkFileListIterator.hasNext()) {
                        File checkFile = (File) checkFileListIterator.next();
                        File transferFile = (File) checkFileList.get(checkFile);
                        if (checkFile == null || transferFile == null) throw new Exception("..error occurred, empty file list is corrupted");
                        
                        // if no files with more than zero byte have been transferred then all zero byte files will be removed
                        if (transferFileList.isEmpty() || zeroByteFilesRelaxed) {
                            this.getLogger().info("removing local temporary file : " + checkFile.getAbsolutePath() + " due to zero byte constraint");
                            if (!checkFile.delete()) {
                                throw new Exception("could not remove temporary file: " + checkFile.getAbsolutePath());
                            }
                            zeroByteCount++;
                            fileZeroByteNotificationBody += transferFile.getName() + "\n";
                            // otherwise rename files with atomic suffixes to their target names
                        } else {
                            if (atomicSuffix != null && atomicSuffix.length() > 0 && checkFile.getAbsolutePath().endsWith(atomicSuffix)) {
                                this.getLogger().info("renaming local temporary file: " + transferFile.getAbsolutePath());
                                if (!checkFile.renameTo(transferFile)) {
                                    throw new Exception("could not rename temporary file [" + checkFile.getCanonicalPath() + "] to: " + transferFile.getAbsolutePath());
                                }
                            }
                            count++;
                            fileNotificationBody += transferFile.getName() + "\n";
                        }
                    }
                }
                
                
                if (zeroByteCount > 0 && fileZeroByteNotificationTo != null && fileZeroByteNotificationTo.length() > 0) {
                    sendMail(fileZeroByteNotificationTo, fileZeroByteNotificationCC, fileZeroByteNotificationBCC, fileZeroByteNotificationSubject, fileZeroByteNotificationBody);
                }
                
                if (count > 0 && fileNotificationTo != null && fileNotificationTo.length() > 0) {
                    sendMail(fileNotificationTo, fileNotificationCC, fileNotificationBCC, fileNotificationSubject, fileNotificationBody);
                }
                
                
                switch (count) {
                case 0: if (zeroByteCount > 0 && zeroByteFilesRelaxed) {
                    this.setStateText("no matching files found, " + zeroByteCount + " zero byte files skipped");
                    this.getLogger().info("no matching files found, " + zeroByteCount + " zero byte files skipped");
                } else if (zeroByteCount > 0 && zeroByteFilesStrict) {
                    throw new Exception("zero byte file(s) found");
                } else if (forceFiles) { 
                    throw new Exception("no matching files found");
                } else {
                    this.setStateText("no matching files found");
                    this.getLogger().info("no matching files found");
                }
                
                if (!forceFiles) {
                    return false;
                } else {
                    rc = (!forceFiles ? true : !zeroByteFilesRelaxed);
                }
                break;
                case 1: this.getLogger().info("1 file received" + ((zeroByteCount > 0) ? ", " + zeroByteCount + " files skipped due to zero byte constraint" : "") ); 
                this.setStateText("1 file received" + ((zeroByteCount > 0) ? ", " + zeroByteCount + " files skipped due to zero byte constraint" : "") );
                rc = true; 
                break;
                default:    this.getLogger().info(count + " files received" + ((zeroByteCount > 0) ? ", " + zeroByteCount + " files skipped due to zero byte constraint" : "") ); 
                this.setStateText(count + " files received" + ((zeroByteCount > 0) ? ", " + zeroByteCount + " files skipped due to zero byte constraint" : "") );
                rc = true; 
                break;
                }
                
                String fileNames="";
                String filePaths="";
                Iterator transferredIterator = transferFileList.iterator();
                while (transferredIterator.hasNext()){
                    File curFile = (File) transferredIterator.next();
                    filePaths += curFile.getAbsolutePath();
                    fileNames += curFile.getName();
                    if (transferredIterator.hasNext()) {
                        filePaths+=";";
                        fileNames+=";";
                    }
                }
                // return the number of transferred files and filenames
                /*
                if (spooler_job.order_queue() != null) {
                    if (spooler_task.order() != null && spooler_task.order().params() != null) {
                        spooler_task.order().params().set_var("ftp_result_files", Integer.toString(count));
                        spooler_task.order().params().set_var("ftp_result_zero_byte_files", Integer.toString(zeroByteCount ));
                        spooler_task.order().params().set_var("ftp_result_filenames", fileNames);
                        spooler_task.order().params().set_var("ftp_result_filepaths", filePaths);
                    }
                } else {
                    if (spooler_task.params() != null){ 
                        spooler_task.params().set_var("ftp_result_files", Integer.toString(count));
                        spooler_task.params().set_var("ftp_result_zero_byte_files", Integer.toString(zeroByteCount));
                        spooler_task.params().set_var("ftp_result_filenames", fileNames);
                        spooler_task.params().set_var("ftp_result_filepaths", filePaths);
                    }
                }
                
                return (spooler_task.job().order_queue() == null) ? false : rc;                             
                */
                return rc;
                
            } catch (Exception e) {
                rc = false;
                this.setStateText("could not process file transfer: " + e.getMessage());
                throw (new Exception("could not process file transfer: " + e.getMessage()));
            } finally {
                if (ftpClient != null) {
                    if (isLoggedIn) try { ftpClient.logout(); } catch (Exception e) {} // no error handling
                    if (ftpClient.isConnected()) try { ftpClient.disconnect(); } catch(Exception e) {} // no error handling
                }
            }
            
        } catch (Exception e){
            this.getLogger().warn ("ftp processing failed: " + e.getMessage());
            return false;
        }
    }
    
    
    public boolean send() throws Exception {
        
        boolean     rc                              = false;       
        SOSFTP      ftpClient                       = null;
        
        String      host                            = "";
        int         port                            = 21;
        
        String      user                            = "anonymous";
        String      password                        = "";
        String      account                         = "";
        
        String      transferMode                    = "binary";
        boolean     passiveMode                     = false;
        
        String      remoteDir                       = "./";
        String      localDir                        = ".";
        String      fileSpec                        = ".*";
        
        String       atomicSuffix                   = "";
        boolean      checkSize                      = true;
        boolean      overwriteFiles                 = true;
        boolean      appendFiles                    = false;
        boolean      removeFiles                    = false;
        boolean      makeDirs                       = false;
        boolean      forceFiles                     = true;
        boolean      compressFiles                  = false;
        String       compressedFileExtension        = "";

        boolean      zeroByteFiles                  = true;
        boolean      zeroByteFilesStrict            = false; 
        boolean      zeroByteFilesRelaxed           = false;
        
        String       fileNotificationTo             = "";
        String       fileNotificationCC             = "";
        String       fileNotificationBCC            = "";
        String       fileNotificationSubject        = "";
        String       fileNotificationBody           = "";
        
        String       fileZeroByteNotificationTo     = "";
        String       fileZeroByteNotificationCC     = "";
        String       fileZeroByteNotificationBCC    = "";
        String       fileZeroByteNotificationSubject= "";
        String       fileZeroByteNotificationBody   = "";
        
        boolean      isLoggedIn                     = false;
        
        boolean      keepConnection                 = false;
        boolean      sameConnection                 = false;
        
        int          count                          = 0;
        int          zeroByteCount                  = 0;
        
        String       replacing                      = null;
        String       replacement                    = null;             
        
        boolean      recursive                      = false;        
        String       filePath                       = "";
        boolean      isFilePath                     = false;                
        
        //alternative Parameter;
        String      alternativeHost                 = "";
        int         alternativePort                 = 0;
        String      alternativeUser                 = ""; 
        String      alternativePassword             = "";
        String      alternativeAccount              = "";
        String      alternativeRemoteDir            = "";
        boolean     alternativePassiveMode          = false;
        String      alternativeTransferMode         = "";

        String      lastHost         = "";
        
        int         lastPort         = 0;
        
        String      lastUser         = "";
        
        String      lastAccount      = "";
        
        
        try {   
            if (this.getLogger() == null) this.setLogger(new SOSStandardLogger(0));
            
            try {
                this.setSettings(new SOSProfileSettings( this.getSettingsFile(), this.getSettingsSection(), this.getLogger()));
            } catch (Exception e) {
                throw new Exception("error occurred retrieving settings: " + e.getMessage());
            }
            
            try { // to get the parameters
                if (settings.getSectionEntry("ftp_keep_connection") != null && settings.getSectionEntry("ftp_keep_connection").length() > 0) {
                    String keep = settings.getSectionEntry("ftp_keep_connection");
                    if (keep.equalsIgnoreCase("true") || keep.equalsIgnoreCase("1")) {
                        keepConnection=true;
                    } else {
                        keepConnection= false;
                    }
                } else {
                    keepConnection=false;
                }
                
                
                if (settings.getSectionEntry("ftp_host") != null && settings.getSectionEntry("ftp_host").length() > 0) 
                    host = settings.getSectionEntry("ftp_host");                  
                
                if (settings.getSectionEntry("ftp_port") != null && settings.getSectionEntry("ftp_port").length() > 0)
                    port = Integer.parseInt(settings.getSectionEntry("ftp_port"));
                
                if (settings.getSectionEntry("ftp_user") != null && settings.getSectionEntry("ftp_user").length() > 0)
                    user = settings.getSectionEntry("ftp_user");
                
                if (settings.getSectionEntry("ftp_password") != null && settings.getSectionEntry("ftp_password").length() > 0)
                    password = settings.getSectionEntry("ftp_password");
                
                if (settings.getSectionEntry("ftp_account") != null && settings.getSectionEntry("ftp_account").length() > 0)
                    account = settings.getSectionEntry("ftp_account");
                
                if (keepConnection){
                    sameConnection = ftpClient!=null && ftpClient.isConnected() && lastHost.equalsIgnoreCase(host)
                    && lastPort==port && lastUser.equalsIgnoreCase(user) && lastAccount.equalsIgnoreCase(account);                      
                }
                if (ftpClient!=null && ftpClient.isConnected() && !sameConnection) {
                    try{
                        if (isLoggedIn) ftpClient.logout();
                        ftpClient.disconnect();
                    } catch (Exception e){}
                }
                
                lastHost = host;
                lastPort = port;
                lastUser = user;
                lastAccount = account;
                
                
                if (settings.getSectionEntry("ftp_transfer_mode") != null && settings.getSectionEntry("ftp_transfer_mode").length() > 0)
                    transferMode = settings.getSectionEntry("ftp_transfer_mode");
                
                if (settings.getSectionEntry("ftp_passive_mode") != null && settings.getSectionEntry("ftp_passive_mode").length() > 0)
                    passiveMode = (settings.getSectionEntry("ftp_passive_mode").equals("1") ? true : false);
                
                if (settings.getSectionEntry("ftp_remote_dir") != null && settings.getSectionEntry("ftp_remote_dir").length() > 0)
                    remoteDir = settings.getSectionEntry("ftp_remote_dir");
                
                if (settings.getSectionEntry("ftp_local_dir") != null && settings.getSectionEntry("ftp_local_dir").length() > 0)
                    localDir = settings.getSectionEntry("ftp_local_dir");
                
                if (settings.getSectionEntry("ftp_file_spec") != null && settings.getSectionEntry("ftp_file_spec").length() > 0)
                    fileSpec = settings.getSectionEntry("ftp_file_spec");            
                
                if (settings.getSectionEntry("ftp_atomic_suffix") != null && settings.getSectionEntry("ftp_atomic_suffix").length() > 0)
                    atomicSuffix = settings.getSectionEntry("ftp_atomic_suffix");
                
                if (settings.getSectionEntry("ftp_check_size") != null && settings.getSectionEntry("ftp_check_size").length() > 0)
                    checkSize = (!settings.getSectionEntry("ftp_check_size").equals("1") 
                            && !settings.getSectionEntry("ftp_check_size").equalsIgnoreCase("true") 
                            && !settings.getSectionEntry("ftp_check_size").equalsIgnoreCase("yes") ? false : true);
                
                if (settings.getSectionEntry("ftp_overwrite_files") != null && settings.getSectionEntry("ftp_overwrite_files").length() > 0)
                    overwriteFiles = (!settings.getSectionEntry("ftp_overwrite_files").equals("1") 
                            && !settings.getSectionEntry("ftp_overwrite_files").equalsIgnoreCase("true") 
                            && !settings.getSectionEntry("ftp_overwrite_files").equalsIgnoreCase("yes") ? false : true);
                
                if (settings.getSectionEntry("ftp_append_files") != null && settings.getSectionEntry("ftp_append_files").length() > 0)
                    appendFiles = (settings.getSectionEntry("ftp_append_files").equals("1") 
                            || settings.getSectionEntry("ftp_append_files").equalsIgnoreCase("true") 
                            || settings.getSectionEntry("ftp_append_files").equalsIgnoreCase("yes") ? true : false);
                
                if (settings.getSectionEntry("ftp_remove_files") != null && settings.getSectionEntry("ftp_remove_files").length() > 0)
                    removeFiles = (settings.getSectionEntry("ftp_remove_files").equals("1") 
                            || settings.getSectionEntry("ftp_remove_files").equalsIgnoreCase("true") 
                            || settings.getSectionEntry("ftp_remove_files").equalsIgnoreCase("yes") ? true : false);
                
                if (settings.getSectionEntry("ftp_make_dirs") != null && settings.getSectionEntry("ftp_make_dirs").length() > 0)
                    makeDirs = (settings.getSectionEntry("ftp_make_dirs").equals("1") 
                            || settings.getSectionEntry("ftp_make_dirs").equalsIgnoreCase("true") 
                            || settings.getSectionEntry("ftp_make_dirs").equalsIgnoreCase("yes") ? true : false);
                
                if (settings.getSectionEntry("ftp_force_files") != null && settings.getSectionEntry("ftp_force_files").length() > 0)
                    forceFiles = (!settings.getSectionEntry("ftp_force_files").equals("1") 
                            && !settings.getSectionEntry("ftp_force_files").equalsIgnoreCase("true") 
                            && !settings.getSectionEntry("ftp_force_files").equalsIgnoreCase("yes") ? false : true);
                
                if (settings.getSectionEntry("ftp_compress_files") != null && settings.getSectionEntry("ftp_compress_files").length() > 0)
                    compressFiles = (settings.getSectionEntry("ftp_compress_files").equals("1") 
                            || settings.getSectionEntry("ftp_compress_files").equalsIgnoreCase("true") 
                            || settings.getSectionEntry("ftp_compress_files").equalsIgnoreCase("yes") ? true : false);
                
                if (settings.getSectionEntry("ftp_compressed_file_extension") != null && settings.getSectionEntry("ftp_compressed_file_extension").length() > 0) {
                    compressedFileExtension = settings.getSectionEntry("ftp_compressed_file_extension");
                } else {
                    compressedFileExtension = ".gz";
                }
                
                
                if (settings.getSectionEntry("ftp_file_zero_byte_transfer") != null && settings.getSectionEntry("ftp_file_zero_byte_transfer").length() > 0) {
                    if (settings.getSectionEntry("ftp_file_zero_byte_transfer").equals("1") 
                            || settings.getSectionEntry("ftp_file_zero_byte_transfer").equalsIgnoreCase("true") 
                            || settings.getSectionEntry("ftp_file_zero_byte_transfer").equalsIgnoreCase("yes")) {
                        zeroByteFiles = true;
                        zeroByteFilesStrict = false;
                    } else if (settings.getSectionEntry("ftp_file_zero_byte_transfer").equalsIgnoreCase("strict")) {
                        zeroByteFiles = false;
                        zeroByteFilesStrict = true;
                    } else if (settings.getSectionEntry("ftp_file_zero_byte_transfer").equalsIgnoreCase("relaxed")) {
                        zeroByteFiles = false;
                        zeroByteFilesStrict = false;
                        zeroByteFilesRelaxed = true;
                    } else {
                        zeroByteFiles = false;
                        zeroByteFilesStrict = false;
                    }
                }
                
                if (settings.getSectionEntry("ftp_file_notification_to") != null && settings.getSectionEntry("ftp_file_notification_to").length() > 0) {
                    fileNotificationTo = settings.getSectionEntry("ftp_file_notification_to");
                }
                
                if (settings.getSectionEntry("ftp_file_notification_cc") != null && settings.getSectionEntry("ftp_file_notification_cc").length() > 0) {
                    fileNotificationCC = settings.getSectionEntry("ftp_file_notification_cc");
                }
                
                if (settings.getSectionEntry("ftp_file_notification_bcc") != null && settings.getSectionEntry("ftp_file_notification_bcc").length() > 0) {
                    fileNotificationBCC = settings.getSectionEntry("ftp_file_notification_bcc");
                }
                
                if (settings.getSectionEntry("ftp_file_notification_subject") != null && settings.getSectionEntry("ftp_file_notification_subject").length() > 0) {
                    fileNotificationSubject = settings.getSectionEntry("ftp_file_notification_subject");
                }
                
                if (settings.getSectionEntry("ftp_file_notification_body") != null && settings.getSectionEntry("ftp_file_notification_body").length() > 0) {
                    fileNotificationBody = settings.getSectionEntry("ftp_file_notification_body");
                }
                
                
                if (settings.getSectionEntry("ftp_file_zero_byte_notification_to") != null && settings.getSectionEntry("ftp_file_zero_byte_notification_to").length() > 0) {
                    fileZeroByteNotificationTo = settings.getSectionEntry("ftp_file_zero_byte_notification_to");
                }
                
                if (settings.getSectionEntry("ftp_file_zero_byte_notification_cc") != null && settings.getSectionEntry("ftp_file_zero_byte_notification_cc").length() > 0) {
                    fileZeroByteNotificationCC = settings.getSectionEntry("ftp_file_zero_byte_notification_cc");
                }
                
                if (settings.getSectionEntry("ftp_file_zero_byte_notification_bcc") != null && settings.getSectionEntry("ftp_file_zero_byte_notification_bcc").length() > 0) {
                    fileZeroByteNotificationBCC = settings.getSectionEntry("ftp_file_zero_byte_notification_bcc");
                }
                
                if (settings.getSectionEntry("ftp_file_zero_byte_notification_subject") != null && settings.getSectionEntry("ftp_file_zero_byte_notification_subject").length() > 0) {
                    fileZeroByteNotificationSubject = settings.getSectionEntry("ftp_file_zero_byte_notification_subject");
                }
                
                if (settings.getSectionEntry("ftp_file_zero_byte_notification_body") != null && settings.getSectionEntry("ftp_file_zero_byte_notification_body").length() > 0) {
                    fileZeroByteNotificationBody = settings.getSectionEntry("ftp_file_zero_byte_notification_body");
                }
                
                if ( settings.getSectionEntry("replacing")!=null && settings.getSectionEntry("replacing").length()>0 ) {                
                    replacing = settings.getSectionEntry("replacing");                  
                }
                
                if ( settings.getSectionEntry("replacement")!=null && settings.getSectionEntry("replacement").length()>0 ) {                
                    replacement = settings.getSectionEntry("replacement");                  
                }               
                
                
                if ( settings.getSectionEntry("ftp_recursive")!=null && settings.getSectionEntry("ftp_recursive").length()>0 ) {    
                    String sRecursive = "";
                    sRecursive = settings.getSectionEntry("ftp_recursive");
                    recursive = sosString.parseToBoolean(sRecursive);
                }   
                
                if ( replacing != null && replacement == null ) {
                    throw new Exception("job parameter is missing for specified parameter [replacing]: [replacement]");
                }
                
                if ( replacing == null && replacement != null ) {
                    throw new Exception("job parameter is missing for specified parameter [replacement]: [replacing]");
                }
                
                
                if (appendFiles && compressFiles) {
                    throw new Exception("unsupported parameter settings [ftp_append_files, ftp_compress_files]: cannot append to compressed files");
                }
                
                if (settings.getSectionEntry("ftp_file_path") != null && settings.getSectionEntry("ftp_file_path").length() > 0) {
                    filePath = settings.getSectionEntry("ftp_file_path");
                    isFilePath = true;
                }
                
                //alternative Parameter
                if (settings.getSectionEntry("ftp_alternative_host") != null && settings.getSectionEntry("ftp_alternative_host").length() > 0) 
                    alternativeHost = settings.getSectionEntry("ftp_alternative_host");                   
                
                if (settings.getSectionEntry("ftp_alternative_port") != null && settings.getSectionEntry("ftp_alternative_port").length() > 0)
                    alternativePort = Integer.parseInt(settings.getSectionEntry("ftp_alternative_port"));
                
                if (settings.getSectionEntry("ftp_alternative_password") != null && settings.getSectionEntry("ftp_alternative_password").length() > 0)
                    alternativePassword = settings.getSectionEntry("ftp_alternative_password");
                
                if (settings.getSectionEntry("ftp_alternative_user") != null && settings.getSectionEntry("ftp_alternative_user").length() > 0)
                    alternativeUser = settings.getSectionEntry("ftp_alternative_user");
                
                if (settings.getSectionEntry("ftp_alternative_account") != null && settings.getSectionEntry("ftp_alternative_account").length() > 0)
                    alternativeAccount = settings.getSectionEntry("ftp_alternative_account");
                
                if (settings.getSectionEntry("ftp_alternative_remote_dir") != null && settings.getSectionEntry("ftp_alternative_remote_dir").length() > 0)
                    alternativeRemoteDir = settings.getSectionEntry("ftp_alternative_remote_dir");
                
                if (settings.getSectionEntry("ftp_alternative_passive_mode") != null && settings.getSectionEntry("ftp_alternative_passive_mode").length() > 0)
                    alternativePassiveMode = (settings.getSectionEntry("ftp_alternative_passive_mode").equals("1") ? true : false);
                
                if (settings.getSectionEntry("ftp_alternative_transfer_mode") != null && settings.getSectionEntry("ftp_alternative_transfer_mode").length() > 0)
                    alternativeTransferMode = settings.getSectionEntry("ftp_alternative_transfer_mode");
                
            } catch (Exception e) {
                throw (new Exception("could not process job parameters: " + e.getMessage()));
            }           
            
            
            if (fileNotificationTo != null && fileNotificationTo.length() > 0) {
                if (fileNotificationSubject == null || fileNotificationSubject.length() == 0) {
                    fileNotificationSubject = "[info] SOSFTPCommand";
                }
                
                if (fileNotificationBody == null || fileNotificationBody.length() == 0) {
                    fileNotificationBody = "The following files have been sent:\n\n";
                }
            }
            
            if (fileZeroByteNotificationTo != null && fileZeroByteNotificationTo.length() > 0) {
                if (fileZeroByteNotificationSubject == null || fileZeroByteNotificationSubject.length() == 0) {
                    fileZeroByteNotificationSubject = "[warning] SOSFTPCommand";
                }
                
                if (fileZeroByteNotificationBody == null || fileZeroByteNotificationBody.length() == 0) {
                    fileZeroByteNotificationBody = "The following files have not been sent and were removed due to zero byte constraints:\n\n";
                }
            }
            
            
            this.getLogger().debug(".. job parameter [ftp_host]                               : " + host);
            this.getLogger().debug(".. job parameter [ftp_port]                               : " + port);
            this.getLogger().debug(".. job parameter [ftp_user]                               : " + user);          
            this.getLogger().debug(".. job parameter [ftp_account]                            : " + account);
            this.getLogger().debug(".. job parameter [ftp_transfer_mode]                      : " + transferMode);
            this.getLogger().debug(".. job parameter [ftp_passive_mode]                       : " + passiveMode);
            this.getLogger().debug(".. job parameter [ftp_remote_dir]                         : " + remoteDir);
            this.getLogger().debug(".. job parameter [ftp_local_dir]                          : " + localDir);
            this.getLogger().debug(".. job parameter [ftp_file_spec]                          : " + fileSpec);
            this.getLogger().debug(".. job parameter [ftp_atomic_suffix]                      : " + atomicSuffix);
            this.getLogger().debug(".. job parameter [ftp_check_size]                         : " + checkSize);
            this.getLogger().debug(".. job parameter [ftp_append_files]                       : " + appendFiles);
            this.getLogger().debug(".. job parameter [ftp_overwrite_files]                    : " + overwriteFiles);
            this.getLogger().debug(".. job parameter [ftp_remove_files]                       : " + removeFiles);
            this.getLogger().debug(".. job parameter [ftp_force_files]                        : " + forceFiles);
            this.getLogger().debug(".. job parameter [ftp_make_dirs]                          : " + makeDirs);
            this.getLogger().debug(".. job parameter [ftp_compress_files]                     : " + compressFiles);
            this.getLogger().debug(".. job parameter [ftp_compressed_file_extension]          : " + compressedFileExtension);
            
            this.getLogger().debug(".. job parameter [ftp_zero_byte_transfer] zeroByte        : " + zeroByteFiles);
            this.getLogger().debug(".. job parameter [ftp_zero_byte_transfer] strict          : " + zeroByteFilesStrict);
            this.getLogger().debug(".. job parameter [ftp_zero_byte_transfer] relaxed         : " + zeroByteFilesRelaxed);
            
            this.getLogger().debug(".. job parameter [ftp_file_notification_to]               : " + fileNotificationTo);
            this.getLogger().debug(".. job parameter [ftp_file_notification_cc]               : " + fileNotificationCC);
            this.getLogger().debug(".. job parameter [ftp_file_notification_bcc]              : " + fileNotificationBCC);
            this.getLogger().debug(".. job parameter [ftp_file_notification_subject]          : " + fileNotificationSubject);
            this.getLogger().debug(".. job parameter [ftp_file_notification_body]             : " + fileNotificationBody);
            
            this.getLogger().debug(".. job parameter [ftp_file_zero_byte_notification_to]     : " + fileZeroByteNotificationTo);
            this.getLogger().debug(".. job parameter [ftp_file_zero_byte_notification_cc]     : " + fileZeroByteNotificationCC);
            this.getLogger().debug(".. job parameter [ftp_file_zero_byte_notification_bcc]    : " + fileZeroByteNotificationBCC);
            this.getLogger().debug(".. job parameter [ftp_file_zero_byte_notification_subject]: " + fileZeroByteNotificationSubject);
            this.getLogger().debug(".. job parameter [ftp_file_zero_byte_notification_body]   : " + fileZeroByteNotificationBody);
            
            this.getLogger().debug(".. job parameter [replacing]                              : " + replacing);
            this.getLogger().debug(".. job parameter [replacement]                            : " + replacement);
            
            this.getLogger().debug(".. job parameter [ftp_recursive]                          : " + recursive);
            
            this.getLogger().debug(".. job parameter [ftp_file_path]                          : " + filePath);
            
            //alternative Parameter
            this.getLogger().debug(".. job parameter [ftp_alternative_host]                   : " + alternativeHost);
            this.getLogger().debug(".. job parameter [ftp_alternative_port]                   : " + alternativePort);
            this.getLogger().debug(".. job parameter [ftp_alternative_user]                   : " + alternativeUser);
            this.getLogger().debug(".. job parameter [ftp_alternative_password]               : " + alternativePassword);
            this.getLogger().debug(".. job parameter [ftp_alternative_account]                : " + alternativeAccount);
            this.getLogger().debug(".. job parameter [ftp_alternative_remote_dir]             : " + alternativeRemoteDir);
            this.getLogger().debug(".. job parameter [ftp_alternative_passive_mode]           : " + alternativePassiveMode);
            this.getLogger().debug(".. job parameter [ftp_alternative_transfer_mode]          : " + alternativeTransferMode);
            
            try { // to check parameters
                if (host == null || host.length() == 0) throw new Exception("no host was specified");
                if (user == null || user.length() == 0) throw new Exception("no user was specified");
                //if (fileSpec == null || fileSpec.length() == 0) throw new Exception("no file specification was given");
                
                if(isFilePath) {
                    if(sosString.parseToString(fileSpec).length() > 0 &&  !fileSpec.equals(".*") ) {
                        // getLogger().warn("parameter [ftp_file_spec] will be ignored because parameter ftp_file_path has been specified. ");
                    }
                    
                    if(sosString.parseToString(remoteDir).length() > 0 &&  !remoteDir.equals("./") ) {
                        // getLogger().warn("parameter [ftp_remote_dir] will be ignored because parameter ftp_file_path has been specified. ");
                    }
                }
            } catch (Exception e) {
                throw (new Exception("invalid or insufficient parameters: " + e.getMessage()));
            }
            
        
            try { // to process ftp
                
                if (!sameConnection){
                    
                    this.getLogger().info("connecting to host " + host + ", port " + port + ", local directory " + localDir + ", remote directory " + remoteDir + 
                            (isFilePath ? ", file " + filePath : ", file specification " + fileSpec) );
                    
                    
                    boolean alternativeUse = true;//Wiederholungskriterien
                    int isAlternativeParameterUse = 0;//Abbruchbedingug, versuch mit alternative Parameter nur einmal
                    while (alternativeUse && isAlternativeParameterUse <= 1) {
                        try {   
//                          initial         
                            ftpClient = new SOSFTP(host, port);
                            this.getLogger().debug("..ftp server reply [init] [host=" + host + "], [port="+ port + "]: " + ftpClient.getReplyString() );
                            
                            //login
                            if (account != null && account.length() > 0) {
                                isLoggedIn = ftpClient.login(user, password, account);                      
                                this.getLogger().debug("..ftp server reply [login] [user=" + user + "], [account=" + account + "]: " + ftpClient.getReplyString() );                        
                            } else {
                                isLoggedIn = ftpClient.login(user, password);
                                this.getLogger().debug("..ftp server reply [login] [user=" + user + "]: " + ftpClient.getReplyString());                        
                            }
                            
                            if (!isLoggedIn || ftpClient.getReplyCode() > ERROR_CODE) {
                                throw new Exception("..ftp server reply [login failed] [user=" + user + "], [account=" + account + "]: " + ftpClient.getReplyString() );
                            }
                            alternativeUse = false;
                            
                        } catch (Exception ex) {
                            
                            this.getLogger().info("..error in ftp server init with [host=" + host + "], [port="+ port + "] " + ex.getMessage());
                            
                            alternativeUse = (alternativeHost.concat(alternativeUser).concat(alternativePassword).
                                    concat(alternativeAccount).length() > 0 || alternativePort != 0);
                            
                            if(alternativeUse && isAlternativeParameterUse == 0) {
                            //Verbindugsaufbau mit alternativen Parameter versuchen 
                                if (ftpClient != null) {
                                    int orderQueueLength = 0;
                                    this.getLogger().debug("..ftp server reply: "  + ftpClient.getReplyString());
                                    
                                    if ((ftpClient.isConnected() && !keepConnection ) ||
                                            (ftpClient.isConnected() && keepConnection && orderQueueLength<1)){
                                        if (isLoggedIn) try { ftpClient.logout(); } catch(Exception e) {} // no error handling
                                        this.getLogger().debug("..ftp server reply [logout]: " + ftpClient.getReplyString());
                                        try { ftpClient.disconnect(); } catch(Exception e) {} // no error handling
                                    }
                                    
                                }
                                                                
                                isAlternativeParameterUse++;
                                host = getAlternative(host, alternativeHost);
                                port = getAlternative(port, alternativePort);
                                user = getAlternative(user, alternativeUser); 
                                password = getAlternative(password, alternativePassword);                       
                                account = getAlternative(account, alternativeAccount);
                                remoteDir = getAlternative(remoteDir, alternativeRemoteDir);
                                passiveMode = alternativePassiveMode;
                                transferMode = getAlternative(transferMode, alternativeTransferMode);
                                this.getLogger().info("..try login with alternative parameter [host=" + host + "], [port="+ port + "] " +
                                        "[user=" + user + "], [account=" + account + "], [remoteDir=" + remoteDir +"], [passiveMode=" + passiveMode +"], " +
                                        "[transferMode=" + transferMode + "]");
                                
                            } else {
                                //wie gehabt Fehler
                                throw new Exception ("..error in ftp server init with [host=" + host + "], [port="+ port + "] " + ex.getMessage());
                            }                           
                            
                        }
                    }
                    
                    
                } else {
                    this.getLogger().info("reusing connection from previous transfer");
                }
                
                //passivemode
                if (passiveMode) {                  
                    ftpClient.passive();                                        
                    if (ftpClient.getReplyCode() > ERROR_CODE) {
                        throw new Exception("..ftp server reply [passive]: "  + ftpClient.getReplyString());
                    } else {
                        this.getLogger().debug("..ftp server reply [passive]: "  + ftpClient.getReplyString());
                    }
                }               
                
                //transfermode
                if (transferMode.equalsIgnoreCase("ascii")) {                   
                    if (ftpClient.ascii()) {
                        this.getLogger().debug("..using ASCII mode for file transfer");
                        this.getLogger().debug("..ftp server reply [ascii]: "  + ftpClient.getReplyString());
                    } else {                                            
                        throw new Exception(".. could not switch to ASCII mode for file transfer ..ftp server reply [ascii]: "  + ftpClient.getReplyString());                                              
                    }
                } else {
                    if (ftpClient.binary()) {                       
                        this.getLogger().debug("using binary mode for file transfer");
                        this.getLogger().debug("..ftp server reply [binary]: "  + ftpClient.getReplyString());
                    } else {
                        throw new Exception(".. could not switch to binary mode for file transfer ..ftp server reply [ascii]: "  + ftpClient.getReplyString());
                        
                    }
                }               
                
                if (makeDirs) {                    
                    //Vector ftpFiles = ftpClient.nList(remoteDir);
                    int cd = ftpClient.cd(remoteDir);
                    if (ftpClient.getReplyCode() <= ERROR_CODE) { // error code signals non-existing directory                       
                        this.getLogger().debug("..ftp server reply [directory exists] [" + remoteDir + "]: "  + ftpClient.getReplyString());
                    //} else if (ftpFiles.isEmpty()) {
                    } else {
                        boolean ok = ftpClient.mkdir(remoteDir);
                        if (!ok || ftpClient.getReplyCode() > ERROR_CODE) {
                            throw new Exception("..error occurred creating directory [" + remoteDir + "]: "  + ftpClient.getReplyString());
                        } else {
                            this.getLogger().debug("..ftp server reply [mkdir] [" + remoteDir + "]: "  + ftpClient.getReplyString());
                        }
                    }
                }
                
                if(remoteDir != null && remoteDir.length() > 0)
                    ftpClient.cd(remoteDir);                
                
                if(ftpClient.getReplyCode() > ERROR_CODE && sosString.parseToString(alternativeRemoteDir).length() > 0) {//alternative Parameter
                    this.getLogger().debug("..ftp server reply [cd] [remoteDir=" + remoteDir + "]: "  + ftpClient.getReplyString());
                    this.getLogger().info("..try with alternative parameter [remoteDir=" + alternativeRemoteDir +"]");
                    ftpClient.cd(alternativeRemoteDir);
                    remoteDir = alternativeRemoteDir;
                }
                
                if (ftpClient.getReplyCode() > ERROR_CODE) {
                    throw new Exception("..ftp server reply [cd] [remoteDir=" + remoteDir + "]: "  + ftpClient.getReplyString());                   
                } else { 
                    this.getLogger().debug("..ftp server reply [cd] [remoteDir=" + remoteDir + "]: "  + ftpClient.getReplyString());
                }
                
                Vector filelist = null;
                if (isFilePath) {                   
                    filelist = new Vector();
                    filelist.add(filePath);     
                    fileSpec = ".*";
                } else {
                    if(new File(localDir).isDirectory()) {              
                        this.getLogger().info("local directory: " + localDir + ", remote directory: " + remoteDir + ", file specification: " + fileSpec);               
                        filelist = SOSFile.getFilelist( new File(localDir).getAbsolutePath(), fileSpec, 0, recursive);
                    } else {
                        this.getLogger().info("local file: " + localDir + ", remote directory: " + remoteDir + ", file specification: " + fileSpec);
                        filelist = new Vector();
                        filelist.add(localDir);
                    }
                }
                Iterator iterator = filelist.iterator();
                
                
                
                if (!zeroByteFiles) { 
                    while(iterator.hasNext()) {
                        File checkFile = (File) iterator.next();
                        if (checkFile.exists()) {
                            if (checkFile.length() == 0) { 
                                this.getLogger().info("skipping transfer of local file: " + checkFile.getAbsolutePath() + " due to zero byte constraint");
                                filelist.remove(checkFile);
                                if (removeFiles) {
                                    if (!checkFile.delete()) {
                                        throw new Exception("..error occurred, could not remove local file: " + checkFile.getAbsolutePath());
                                    } else {
                                        this.getLogger().info("removing file: " + checkFile.getAbsolutePath());
                                    }
                                }
                                zeroByteCount++;
                                fileZeroByteNotificationBody += checkFile.getName() + "\n";
                            }
                        }
                    }
                }
                
                iterator = filelist.iterator();
                while(iterator.hasNext()) {
                    long bytesSend = 0;
                    Object fn = iterator.next();
                    File localFile = new File(fn.toString());
                    File subParent = null;
                    String subPath = "";
                    
                    if(!localFile.exists()) {
                        throw new Exception(".. file [" + localFile + "] does not exist ");
                    }
                    
                    if(recursive && !isFilePath) {
                        //berprfen, ob das Verzeichnis auf den FTP Server existiert, wenn nicht dann soll das gleiche Verzeichnis generiert werden
                        if(localFile.getParent() != null && localFile.getParentFile().isDirectory()) {//es existieren Vaterknoten                       
                            subPath = fn.toString().substring((localDir.length() + 1));//Unterverzeichnisse sind alle Verzeichnisse unterhalb der localDir                          
                            subParent = new File(subPath).getParentFile();
                            
                            if(subParent != null) {
                                
                                subPath = subPath.replaceAll("\\\\", "/");
                                subPath = subPath.substring(0, subPath.length() - new File(fn.toString()).getName().length() - 1) ;
                                this.getLogger().debug4("..subpath [" + subPath + "]");                             
                                String[] ftpFiles = ftpClient.listNames(remoteDir+ "/" + subPath);                              
                                if (ftpFiles.length == 0) {
                                
                                    boolean ok = ftpClient.mkdir(remoteDir + "/" + subPath);
                                    if (!ok || ftpClient.getReplyCode() > ERROR_CODE) {
                                        throw new Exception("..error occurred creating sub directory [" + remoteDir + "/" + subPath + "]: "  + ftpClient.getReplyString());
                                    } else {
                                        this.getLogger().debug("..ftp server reply [mkdir sub directory] [" + remoteDir + "/" + subPath + "]: "  + ftpClient.getReplyString());
                                    }
                                }
                                
                            }
                        }
                        
                    }
                    
                    if (localFile == null) continue;
                    if (!zeroByteFiles && localFile.length() == 0) continue;
                    
                    File transferFile = null;
                    String transferFilename = null; //Hilfsvariable 
                    File sourceFile = null;
                    if (compressFiles) {
                        transferFile = new File(localFile.getAbsolutePath() + compressedFileExtension);
                        sourceFile = File.createTempFile("sos", ".gz");
                    } else {
                        transferFile = new File(localFile.getAbsolutePath());
                        sourceFile = new File(localFile.getAbsolutePath());
                    }
                    
                    transferFilename = transferFile.getName();
                    if (replacement != null && replacement.length() > 0 && replacing != null && replacing.length() > 0) {
                        String currTransferFilename = SOSFileOperations.getReplacementFilename(transferFile.getName(), replacing, replacement);
                        this.getLogger().info("source filename ["+transferFile.getName()+"] is renamed to: " + currTransferFilename);
                        transferFile = new File(currTransferFilename);
                        transferFilename = transferFile.getName();
                    }
                    
                    if(subParent != null && recursive) {
                        transferFile = new File((subParent != null ? subParent.getName() + "/" : "") + transferFile.getName());
                        transferFilename = (subParent != null ? subPath + "/" : "") + transferFile.getName();
                    }
                    
                    
                    if (!appendFiles && atomicSuffix != null && atomicSuffix.length() > 0) {
                        if (!overwriteFiles) {
                            Vector ftpFiles = ftpClient.nList(remoteDir + (remoteDir.endsWith("/") || remoteDir.endsWith("\\") ? "" : "/") + transferFile.getName());
                            if (ftpClient.getReplyCode() <= ERROR_CODE) { // error code signals non-existing file
                                this.getLogger().debug("..ftp server reply [file exists] [" + transferFilename + "]: "  + ftpClient.getReplyString());
                                if (!ftpFiles.isEmpty()) {
                                    this.getLogger().info("..ftp transfer skipped for file [no overwrite]: " + transferFilename);
                                    continue;
                                }
                            }
                        }
                        
                        if (compressFiles) {
                            SOSGZip.compressFile(localFile, sourceFile);
                            this.getLogger().info("sending file : " + transferFile.getAbsolutePath() + atomicSuffix + " from temporary file " + sourceFile.getAbsolutePath());
                        } else {
                            this.getLogger().info("sending file : " + transferFile.getAbsolutePath() + atomicSuffix);
                        }                                                
                        
                        
                        java.io.OutputStream outputStream = ftpClient.storeFileStream(transferFilename + atomicSuffix);
                        if (ftpClient.getReplyCode() > 300) {
                            throw new Exception("..error occurred storeFileStream on the Ftp-Server for file [" + transferFilename + atomicSuffix+ "]: "  + ftpClient.getReplyString());
                        } else {
                            this.getLogger().debug("..ftp server reply [storeFileStream] [" + transferFilename + atomicSuffix+ "]: "  + ftpClient.getReplyString());
                        }
                        
                        bytesSend = ftpClient.putFile(sourceFile.getAbsolutePath(), outputStream);
                        
                        if (ftpClient.getReplyCode() <= 300) {
                            if (overwriteFiles) {
                                Vector ftpFiles = ftpClient.nList(transferFilename);
                                if (ftpClient.getReplyCode() <= ERROR_CODE) { // error code signals non-existing file
                                    this.getLogger().debug("..ftp server reply [file exists] [" + transferFilename + "]: "  + ftpClient.getReplyString());
                                    if (!ftpFiles.isEmpty() && ftpFiles.contains(transferFilename)) {
                                        boolean ok = ftpClient.delete(transferFilename);
                                        if (!ok || ftpClient.getReplyCode() > ERROR_CODE) {
                                            throw new Exception("..error occurred overwriting file [" + transferFilename + "]: "  + ftpClient.getReplyString());
                                        } else {
                                            this.getLogger().debug("..ftp server reply [delete] [" + transferFilename + "]: "  + ftpClient.getReplyString());
                                        }
                                    }
                                }
                            }
                            this.getLogger().info("renaming file: " + transferFilename);
                            ftpClient.rename(transferFilename + atomicSuffix, transferFilename);
                        }
                    } else {
                        if (!overwriteFiles && !appendFiles) {
                            Vector ftpFiles = ftpClient.nList(transferFilename);
                            if (ftpClient.getReplyCode() <= ERROR_CODE) { // error code signals non-existing file
                                this.getLogger().debug("..ftp server reply [file exist] [" + transferFilename + "]: "  + ftpClient.getReplyString());
                                /*if (!ftpFiles.isEmpty()) {                                    
                                    this.getLogger().info("..ftp transfer skipped for file [no overwrite]: " + transferFilename);
                                    continue;
                                }*/
                                boolean bWhileContinue = false;
                                for (int i=0; i < ftpFiles.size(); i++ ) {
                                    String currFtpFilename = sosString.parseToString(ftpFiles.get(i));
                                    if(currFtpFilename.equalsIgnoreCase(transferFilename)) {
                                        this.getLogger().info("..ftp transfer skipped for file [no overwrite]: " + transferFilename);
                                        bWhileContinue = true;
                                        continue;                                       
                                    }
                                }
                                if(bWhileContinue)
                                    continue;
                            }
                        }
                        
                        
                        if (compressFiles) {
                            SOSGZip.compressFile(localFile, sourceFile);
                        }
                        
                        if (appendFiles) {
                            this.getLogger().info("sending file : " + transferFile.getAbsolutePath() + " (append)");
                            bytesSend = ftpClient.appendFile( sourceFile.getAbsolutePath(), transferFilename );
                        } else {
                            if (compressFiles) {
                                this.getLogger().info("sending file : " + transferFile.getAbsolutePath() + " from temporary file " + sourceFile.getAbsolutePath());
                            } else {
                                this.getLogger().info("sending file : " + transferFile.getAbsolutePath());
                            }
                            
                            java.io.OutputStream outputStream = ftpClient.storeFileStream(transferFilename);
                            if (ftpClient.getReplyCode() > 300) {
                                throw new Exception("..error occurred storeFileStream on the Ftp-Server for file [" + transferFilename +"]: "  + ftpClient.getReplyString());
                            } else {
                                this.getLogger().debug("..ftp server reply [storeFileStream] [" + transferFilename + "]: "  + ftpClient.getReplyString());
                            }
                            
                            bytesSend = ftpClient.putFile(sourceFile.getAbsolutePath(), outputStream);
                                                        
                        }
                    }                    
                    
                    if (ftpClient.getReplyCode() > ERROR_CODE) {
                        throw new Exception("..error occurred sending file [" + transferFile.getAbsolutePath() + "]: "  + ftpClient.getReplyString());
                    } else {
                        this.getLogger().debug("..ftp server reply [put] [" + transferFile.getAbsolutePath() + ", size=" + bytesSend + "]: "  + ftpClient.getReplyString());
                    }
                    
                    if (checkSize && transferFile.length() > 0 && transferFile.length() != bytesSend) {
                        throw new Exception("..error occurred sending file, local file size [" + transferFile.length() + "] does not match remote file size ["  + bytesSend + "]");
                    }
                    
                    
                    if (compressFiles) {
                        if (!sourceFile.delete()) {
                            throw new Exception("..error occurred, could not remove compressed local file: " + sourceFile.getAbsolutePath());
                        } else {
                            this.getLogger().debug1("removing compressed file: " + sourceFile.getAbsolutePath());
                        }
                    }
                    
                    if (removeFiles) {
                        if (!localFile.delete()) {
                            throw new Exception("..error occurred, could not remove local file: " + localFile.getAbsolutePath());
                        } else {
                            this.getLogger().info("removing file: " + localFile.getAbsolutePath());
                        }
                    }
                    
                    fileNotificationBody += transferFile.getName() + "\n";
                    count++;
                }
                
                
                if (zeroByteCount > 0 && fileZeroByteNotificationTo != null && fileZeroByteNotificationTo.length() > 0) {
                    sendMail(fileZeroByteNotificationTo, fileZeroByteNotificationCC, fileZeroByteNotificationBCC, fileZeroByteNotificationSubject, fileZeroByteNotificationBody);
                }
                
                if (count > 0 && fileNotificationTo != null && fileNotificationTo.length() > 0) {
                    sendMail(fileNotificationTo, fileNotificationCC, fileNotificationBCC, fileNotificationSubject, fileNotificationBody);
                }
                
                
                switch (count) {
                case 0:     if (zeroByteCount > 0 && zeroByteFilesRelaxed) {
                    this.setStateText("no matching files found, " + zeroByteCount + " zero byte files skipped");
                    this.getLogger().info("no matching files found, " + zeroByteCount + " zero byte files skipped");
                } else if (zeroByteCount > 0 && zeroByteFilesStrict) {
                    throw new Exception("zero byte file(s) found");
                } else if (forceFiles) { 
                    throw new Exception("no matching files found");
                } else {
                    this.setStateText("no matching files found");
                    this.getLogger().info("no matching files found");
                }
                rc = (!forceFiles ? true : !zeroByteFilesRelaxed); 
                break;
                case 1:     this.getLogger().info("1 file transferred" + ((zeroByteCount > 0) ? ", " + zeroByteCount + " files skipped due to zero byte constraint" : "") ); 
                this.setStateText("1 file transferred" + ((zeroByteCount > 0) ? ", " + zeroByteCount + " files skipped due to zero byte constraint" : "") );
                rc = true; 
                break;
                default:    this.getLogger().info(count + " files transferred" + ((zeroByteCount > 0) ? ", " + zeroByteCount + " files skipped due to zero byte constraint" : "") ); 
                this.setStateText(count + " files transferred" + ((zeroByteCount > 0) ? ", " + zeroByteCount + " files skipped due to zero byte constraint" : "") );
                rc = true; 
                break;
                }
                
                
                // return the number of transferred files
                /*
                if (spooler_job.order_queue() != null) {
                    spooler_task.order().params().set_var("ftp_result_files", Integer.toString(count));
                    spooler_task.order().params().set_var("ftp_result_zero_byte_files", Integer.toString(zeroByteCount ));
                } else {
                    spooler_task.params().set_var("ftp_result_files", Integer.toString(count));
                    spooler_task.params().set_var("ftp_result_zero_byte_files", Integer.toString(zeroByteCount));
                }
                */

                return rc;
                
            } catch (Exception e) {
                rc = false;
                throw (new Exception("could not process file transfer: " + e.getMessage()));
            } finally {
                if (ftpClient != null) {
                    int orderQueueLength = 0;
                    this.getLogger().debug("..ftp server reply: "  + ftpClient.getReplyString());
                    
                    if ((ftpClient.isConnected() && !keepConnection ) ||
                            (ftpClient.isConnected() && keepConnection && orderQueueLength<1)){
                        if (isLoggedIn) try { ftpClient.logout(); } catch(Exception e) {} // no error handling
                        this.getLogger().debug("..ftp server reply [logout]: " + ftpClient.getReplyString());
                        try { ftpClient.disconnect(); } catch(Exception e) {} // no error handling
                    }
                    
                }
            }
            
        } catch (Exception e){
            this.setStateText("ftp processing failed: " + e.getMessage());
            this.getLogger().warn ("ftp processing failed: " + e.getMessage());
            return false;
        }
    }
    
    
    public boolean remove() throws Exception {
        
        boolean      rc                              = false;       
        SOSFTP       ftpClient                       = null;
        
        return rc;
    }
    
    
    private long transferFile(SOSFTP ftpClient, File sourceFile, File targetFile, long checkRetry, long checkInterval, 
            boolean appendFiles) throws Exception {
        
        long retry = (checkRetry > 0) ? checkRetry : 0;
        long interval = (checkInterval > 0) ? checkInterval : 60;
        long currentBytesSent = 0;
        long previousBytesSent = 0;
                
        
        try {
            this.getLogger().info("receiving file: " + targetFile.getAbsolutePath() + " " + ftpClient.size(sourceFile.getName()) + " bytes");
            
            for(int i=-1; i<retry; i++) {
                currentBytesSent = ftpClient.getFile( sourceFile.getName(), targetFile.getAbsolutePath(), appendFiles );
                
                if (currentBytesSent < 0) currentBytesSent = 0;
                if (ftpClient.getReplyCode() > ERROR_CODE) {
                    throw new Exception("..error occurred receiving file [" + targetFile.getAbsolutePath() + "]: "  + ftpClient.getReplyString());
                } else {
                    this.getLogger().debug("..ftp server reply [getFile] [" + targetFile.getAbsolutePath() + ", size=" + currentBytesSent + "]: " + ftpClient.getReplyString());
                }
                
                if (appendFiles == false && retry > 0 && currentBytesSent != previousBytesSent) {
                    this.getLogger().info("..retry " + (i+2) + " of " + retry + " to wait " + interval + "s for file transfer being completed, current file size: " + currentBytesSent + " bytes");
                    try { 
                        Thread.sleep(checkInterval*1000);
                    } catch (InterruptedException e){} // the VM doesn't want us to sleep anymore, so get back to work
                } else {
                    break;
                }
                
                previousBytesSent = currentBytesSent;
            }
            
            if (targetFile.length() > 0 && ftpClient.size(sourceFile.getName()) != currentBytesSent) {
                throw new Exception("..error occurred receiving file, local file size [" + targetFile.length() + "] does not match remote file size ["  + currentBytesSent + "]");
            }
            
            return currentBytesSent;
        } catch (Exception e){
            throw new Exception("file transfer failed: " + e.getMessage());
        }
        
    }
    
    
    /**
     * send mail with Job Scheduler settings
     * 
     * @param recipient
     * @param recipientCC carbon copy recipient
     * @param recipientBCC blind carbon copy recipient
     * @param subject
     * @param body
     * @throws Exception
     */
    public void sendMail(String recipient, String recipientCC, String recipientBCC, String subject, String body) throws Exception {
        
        try {
            SOSMail sosMail = new SOSMail(this.getMailSMTP());
            
            sosMail.setQueueDir(this.getMailQueueDir());
            sosMail.setFrom(this.getMailFrom());
            sosMail.setContentType("text/plain");
            sosMail.setEncoding("Base64");
            
            String recipients[] = recipient.split(",");
            for(int i=0; i<recipients.length; i++) {
                sosMail.addRecipient(recipients[i].trim());
            }
            
            String recipientsCC[] = recipientCC.split(",");
            for(int i=0; i<recipientsCC.length; i++) {
                sosMail.addCC(recipientsCC[i].trim());
            }
            
            String recipientsBCC[] = recipientBCC.split(",");
            for(int i=0; i<recipientsBCC.length; i++) {
                sosMail.addBCC(recipientsBCC[i].trim());
            }
            
            sosMail.setSubject(subject);
            sosMail.setBody(body);
            sosMail.setSOSLogger(this.getLogger());
            
            this.getLogger().info("sending mail: \n" + sosMail.dumpMessageAsString());
            
            if (!sosMail.send()){
                this.getLogger().warn("mail server is unavailable, mail for recipient [" + recipient + "] is queued in local directory [" + sosMail.getQueueDir() + "]:" + sosMail.getLastError());
            }
            
            sosMail.clearRecipients();
        } catch (Exception e) {
            throw new Exception("error occurred sending mai: " + e.getMessage());
        }
    }   
    
    private String getAlternative(String param, String alternativeParam) throws Exception {
        try {
            if(sosString.parseToString(alternativeParam).length() > 0) {
                return alternativeParam;
            } else {
                return param;
            } 
        } catch (Exception e) {
            this.getLogger().warn("error in getAlternative: "   + e.getMessage());
            return param;
        }
    }
    
    private int getAlternative(int param, int alternativeParam) throws Exception {
        try {
            if(alternativeParam > 0) {
                return alternativeParam;
            } else {
                return param;
            } 
        } catch (Exception e) {
            this.getLogger().warn("error in getAlternative: "   + e.getMessage());
            return param;
        }
    }


    /**
     * @return Returns the logger.
     */
    public SOSLogger getLogger() {
        return logger;
    }


    /**
     * @param logger The logger to set.
     */
    public void setLogger(SOSLogger logger) {
        this.logger = logger;
    }


    /**
     * @return Returns the settingsFile.
     */
    public String getSettingsFile() {
        return settingsFile;
    }


    /**
     * @param settingsFile The settingsFile to set.
     */
    public void setSettingsFile(String settingsFile) {
        this.settingsFile = settingsFile;
    }


    /**
     * @return Returns the state.
     */
    public boolean isState() {
        return state;
    }


    /**
     * @param state The state to set.
     */
    public void setState(boolean state) {
        this.state = state;
    }


    /**
     * @return Returns the stateText.
     */
    public String getStateText() {
        return stateText;
    }


    /**
     * @param stateText The stateText to set.
     */
    public void setStateText(String stateText) {
        this.stateText = stateText;
    }


    /**
     * @return Returns the mailFrom.
     */
    public String getMailFrom() {
        return mailFrom;
    }


    /**
     * @param mailFrom The mailFrom to set.
     */
    public void setMailFrom(String mailFrom) {
        this.mailFrom = mailFrom;
    }


    /**
     * @return Returns the mailQueueDir.
     */
    public String getMailQueueDir() {
        return mailQueueDir;
    }


    /**
     * @param mailQueueDir The mailQueueDir to set.
     */
    public void setMailQueueDir(String mailQueueDir) {
        this.mailQueueDir = mailQueueDir;
    }


    /**
     * @return Returns the mailSMTP.
     */
    public String getMailSMTP() {
        return mailSMTP;
    }


    /**
     * @param mailSMTP The mailSMTP to set.
     */
    public void setMailSMTP(String mailSMTP) {
        this.mailSMTP = mailSMTP;
    }

    /**
     * @return Returns the settings.
     */
    public SOSSettings getSettings() {
        return settings;
    }

    /**
     * @param settings The settings to set.
     */
    public void setSettings(SOSSettings settings) {
        this.settings = settings;
    }

    /**
     * @return Returns the settingsSection.
     */
    public String getSettingsSection() {
        return settingsSection;
    }

    /**
     * @param settingsSection The settingsSection to set.
     */
    public void setSettingsSection(String settingsSection) {
        this.settingsSection = settingsSection;
    }


    public static void main(String[] args) {

        boolean rc = false;
        int logLevel = 0;
        String settingsFile = "";
        String settingsSection = "";
        String operation = "";
        SOSFTPCommand ftpCommand = null;

        String usage = "Usage: sos.net.SOSFTPCommand -operation= -settings= -section= -verbose=" + "\n" +
                       "        -operation=[send|receive|remove]   FTP operation" + "\n" +
                       "        -settings=[file]                   Configuration file" + "\n" +
                       "        -section=[section]                 Section for FTP settings" + "\n" + 
                       "                                           in configuration file" + "\n" + 
                       "        -verbose=[1..9]                    Verbosity level" + "\n";

        try {
            
            SOSArguments arguments = new SOSArguments(args);
            
            operation       = arguments.as_string("-operation=", "");
            settingsFile    = arguments.as_string("-settings=", "");
            settingsSection = arguments.as_string("-section=", "ftp");
            logLevel        = arguments.as_int("-verbose=", SOSLogger.INFO);
            
            try {
                arguments.check_all_used();
            } catch (Exception e) {
                System.out.println(usage);
                throw new Exception("[ERROR] invalid command line parameters specified" + e.getMessage());
            }
            
            SOSStandardLogger logger = new SOSStandardLogger(logLevel);
            ftpCommand = new SOSFTPCommand(settingsFile, settingsSection, logger);
            
            if (operation.equalsIgnoreCase("send")) {
              rc = ftpCommand.send();
            } else if (operation.equalsIgnoreCase("receive")) {
                  rc = ftpCommand.receive();
            } else if (operation.equalsIgnoreCase("delete") || operation.equalsIgnoreCase("remove")) {
                rc = ftpCommand.remove();
            } else {
                System.out.println(usage);
                throw new Exception("[ERROR] no valid operation was specified, use send|receive|remove: " + operation);
            }

            System.exit( (rc ? 0 : 1) );
        } catch (Exception e) {
            System.out.println(e.getMessage());
            System.exit(1);
        }

    }

}


